@@ -24,19 +24,21 @@
 
 #include <sqlite3.h>
 
-// NGram parameters (default)
-#define XAPIAN_DEFAULT_PARTIAL 3L
-#define XAPIAN_DEFAULT_FULL 20L
+#define XAPIAN_SLEEP std::chrono::milliseconds(200)
+
+#define XAPIAN_PLUGIN_VERSION "1.8.4"
 
 // Ressources limits
 #define XAPIAN_FILE_PREFIX "xapian-indexes" // Locations of indexes
 #define XAPIAN_TERM_SIZELIMIT 245L // Hard limit of Xapian library
 #define XAPIAN_MAXTERMS_PERDOC 50000L // Nb of keywords max per email
-#define XAPIAN_WRITING_CACHE 4000L // Max nb of emails processed in cache 
-#define XAPIAN_MIN_RAM 500L // MB
+#define XAPIAN_WRITING_CACHE 5000L // Max nb of emails processed in cache 
+#define XAPIAN_DICT_MAX 60000L // Max nb of terms  in the dict
+#define XAPIAN_MIN_RAM 300L // MB
 #define XAPIAN_MAX_ERRORS 1024L 
 
 // Word processing
+#define XAPIAN_DEFAULT_PARTIAL 3L
 #define XAPIAN_WILDCARD "wldcrd"
 #define XAPIAN_EXPUNGE_HEADER 9
 
@@ -45,24 +47,37 @@
 #define HDRS_NB 11
 static const char * hdrs_emails[HDRS_NB] = { "uid", "subject", "from", "to",  "cc",  "bcc",  "messageid", "listid", "body", "contenttype", ""  };
 static const char * hdrs_xapian[HDRS_NB] = { "Q",   "S",       "A",    "XTO", "XCC", "XBCC", "XMID",      "XLIST",  "XBDY", "XCT", "XBDY" };
-static const char * createTable = "CREATE TABLE IF NOT EXISTS docs(ID INT PRIMARY KEY NOT NULL);";
-static const char * selectUIDs = "select ID from docs;";
+
+static const char * createExpTable = "CREATE TABLE IF NOT EXISTS expunges(ID INTEGER PRIMARY KEY NOT NULL);";
+static const char * selectExpUIDs = "select ID from expunges;";
+static const char * replaceExpUID = "replace into expunges values (%d);";
+static const char * deleteExpUID = "delete from expunges where ID=%d;";
+static const char * suffixExp = "_exp.db";
+
+static const char * createDictTable = "CREATE TABLE IF NOT EXISTS dict (keyword TEXT UNIQUE COLLATE NOCASE, len INTEGER ); CREATE INDEX IF NOT EXISTS dict_len ON dict (len);";
+static const char * createTmpTable = "ATTACH DATABASE ':memory:' AS work; CREATE TABLE work.dict (keyword TEXT UNIQUE COLLATE NOCASE, len INTEGER );";
+static const char * replaceTmpWord ="INSERT OR IGNORE INTO work.dict VALUES('";
+static const char * flushTmpWords = "BEGIN TRANSACTION; INSERT OR IGNORE INTO main.dict SELECT keyword, len FROM work.dict; DELETE FROM work.dict; COMMIT;";
+static const char * searchDict1 = "SELECT keyword FROM dict WHERE (keyword like '%";
+static const char * searchDict2 = "%') ORDER BY len LIMIT 100";
+static const char * suffixDict = "_dict.db";
+
 #define CHAR_KEY "_"
 #define CHAR_SPACE " "
 
 #define CHARS_PB 16
 static const char * chars_pb[] = { "<", ">", ".", "-", "@", "&", "%", "*", "|", "`", "#", "^", "\\", "'", "/", "~" };
 
-#define CHARS_SEP 13
-static const char * chars_sep[] = { "\"", "\r", "\n", "\t", ",", ":", ";", "(", ")", "?", "!", "¿", "¡" };
+#define CHARS_SEP 16
+static const char * chars_sep[] = { "\"", "\r", "\n", "\t", ",", ":", ";", "(", ")", "?", "!", "¿", "¡", "\u00A0", "‘", "“" };
 
 
 struct fts_xapian_settings
 {
 	long verbose;
 	long lowmemory;
-	long partial,full;
-	bool detach;
+	long partial;
+	long maxthreads;
 };
 
 struct fts_xapian_user {
